/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   attachments viewer
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "attach.h"
#include "ecurses.h"
#include "error.h"
#include "mail.h"
#include "rmime.h"
#include "select.h"
#include "xmalloc.h"
#include "read.h"
#include "cmd.h"
#include "gettext.h"
#include "wrapbox.h"
#include "file.h"
#include "ask.h"
#include "interface.h"
#include "color.h"
#include "label.h"
#include "folder.h"
#include "eprintf.h"
#include "mailreader.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/* Attachment window consists of select_t object and an optional label. */
static select_t *attach_select = NULL;
static elabel_t *label         = NULL;


static char *attach_fmt = "%020t %040f (%-06s)";


/* Color used in attachments window. */
static chtype text_color;


/* Used in draw_line. */
static str_t *str_line = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

int attach_selected_item = -1;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

/* This file is generated by interface.pl script from interface.desc,
   and inc.in. */
static WINDOW *interface_init (void);
#include "attach.inc"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
draw_line (WINDOW *win, int maxlen, int index)
{
        mail_t  *mail  = folder_mail_selected ();
        rmime_t *rmime = (mail) ? mail->rmime : NULL;
        mime_t  *mime  = (rmime && rmime->count > index)
                         ? rmime->array[index] : NULL;
        
        if (str_line == NULL)
                str_line = str_create ();
        
        if (mime) {
                eprintf_mime_str (attach_fmt, mime, str_line);
                maxlen -= window_addnstr (win, str_line->str, maxlen);
        }

        while (maxlen--)
                window_addch (win, ' ');
}



static int
count (select_t *nothing)
{
        mail_t *mail = folder_mail_selected ();

        if (mail == NULL || mail->rmime == NULL)
                return 0;
        return mail->rmime->count;
}



static void
set_focus (void)
{
        if (label){
                label_set_focus (label);
        }
        
        cmd_state_push (CMD_ATTACH);
        attach_refresh ();
}



static void
unset_focus (void)
{
        if (label){
                label_unset_focus (label);
                label_redraw (label);
        }

        cmd_state_pop ();
}

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

void
attach_init (void)
{
        WINDOW *window;
        
        window        = interface_init ();
        attach_select = select_open (window, 0, draw_line, count);

        window_set_functions (window, attach_show, attach_refresh,
                              set_focus, unset_focus);
}



void
attach_free_resources (void)
{
        if (attach_select)
                select_close (attach_select);
        attach_select = NULL;

        if (label)
                label_destroy (label);
        label = NULL;

        if (str_line)
                str_destroy (str_line);
        str_line = NULL;
}



void
attach_show (void)
{
        window_show (attach_select->win);
        select_first (attach_select);
        attach_refresh ();
}


void
attach_hide (void)
{
        window_hide (attach_select->win);
}



void
attach_refresh (void)
{
        if (label)
                label_redraw (label);
        select_redraw (attach_select);
}



void
attach_hit (void)
{
        window_hide (attach_select->win);
        mailreader_change_mime (attach_select->bar_pos);
}



void
attach_prev (void)
{
        select_prev (attach_select);
}



void
attach_next (void)
{
        select_next (attach_select);
}



void
attach_prev_page (void)
{
        select_prev_page (attach_select);
}



void
attach_next_page (void)
{
        select_next_page (attach_select);
}



void
attach_first (void)
{
        select_first (attach_select);
}



void
attach_last (void)
{
        select_last (attach_select);
}




void
attach_save (void)
{
        FILE   *fp;
        mime_t *mime;
        char   *file;
        str_t  *str;
        int     ret;
        mail_t *mail = folder_mail_selected ();

        if (mail == NULL)
                return;
  
        if (mail && mail->rmime){
                mime = mail->rmime->array[attach_select->bar_pos];
        }
        else{
                return;
        }

        file = read_argument (_("file: "), mime->file_name, COMPLETE_FILES);
        if (file == NULL || *file == '\0' || *file == '\n')
                return;

        fp = file_open (file, "w", O_WRONLY | O_CREAT | O_EXCL, 0644);
        if (fp == NULL){
                if (errno != EEXIST){
                        error_ (errno, "%s", file);
                        return;
                }
                file = xstrdup (file);

                ret = ask_if_sure (_("file already exists, overwrite [y/N]? "));
    
                switch (ret){

                        case -1:
                        case 0:
                                xfree (file);
                                return;
      
                        case 1:
                                fp = file_open (file, "w", O_WRONLY | O_CREAT | O_TRUNC, 0644);
                                if (fp == NULL){
                                        error_ (errno, "%s", file);
                                        xfree (file);
                                        return;
                                }
                                xfree (file);
                }
        }
  
        str = wrapbox_mail_body (mail, attach_select->bar_pos);
        if (str == NULL){
                error_ (0, _("couldn't get attachment, write aborted"));
                return;
        }
  
        if (fwrite (str->str, 1, str->len, fp) != str->len){
                error_ (errno, "fwrite");
        }
        fclose (fp);

        str_destroy (str);
}



/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE attach.c
 *
 ****************************************************************************/
