/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003, 2004 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   Functions resposible for preparing message headers, calling an editor, etc.

   
*/

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_LOCALE_H
# include <locale.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include "ecurses.h"
#include "xmalloc.h"
#include "mail.h"
#include "rstring.h"
#include "ask.h"
#include "error.h"
#include "file.h"
#include "folder.h"
#include "compose.h"
#include "sender.h"
#include "eprintf.h"
#include "wrapbox.h"
#include "misc.h"
#include "read.h"
#include "str.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#ifdef NO_OFFENSIVE_BOUNDARY
# define BOUNDARY_DELIMITER_FORMAT "--some_elmo_boundary%d"
#else
# define BOUNDARY_DELIMITER_FORMAT "--java_obsysa%d"
#endif

#define DEFAULT_REPLY_FMT "On %d, %f wrote:"
#define LINE_MAX_LEN      70

#define PREAMBLE do {if (mail == NULL) return NULL; } while (0)

typedef enum {COMPOSE_NEW, COMPOSE_REPLY, COMPOSE_FWD} compose_act_t;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/**
 * This indicates what kind of action did we take.  It's set in one of:
 * compose_new, compose_reply, compose_fwd, and used in compose_after_send.
 */
static compose_act_t compose_act = 0;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static char *
message_id (void)
{
        size_t  hlen  = 100;
        char   *hname = xmalloc (hlen);
        char   *result;

        while (gethostname (hname, hlen)){
                hlen  = (hlen + 1) * 2;
                hname = xrealloc (hname, hlen);
        }
        hlen = strlen (hname);

        /*                <  elmo time  pid rand  @ hostname  >  \0 */
        result = xmalloc (1 + 4 +  10 + 10 + 10 + 1 + hlen +  1 + 1);

        sprintf (result, "<elmo%d%d%d@%s>", (int) time (NULL), getpid (),
                 rand (), hname);
        xfree (hname);

        return result;
}


static char *
boundary (void)
{
        char *result;

        result = xmalloc (11 + 10 + 1);

        sprintf (result, BOUNDARY_DELIMITER_FORMAT, rand ());
        return result;
}


static char *
make_date (void)
{
        char        *result;
  
#ifdef HAVE_LOCALE_H
        setlocale (LC_ALL, "C");
#endif

        result = date_string ("%a, %d %b %Y %H:%M:%S %z", time (NULL));

#ifdef HAVE_LOCALE_H
        setlocale (LC_ALL, "");
#endif

        return result;
}



static char *
make_in_reply_to (char *msg_id)
{
        size_t  msg_id_len = strlen (msg_id);
        char   *result;

        result = xmalloc (13 + msg_id_len + 1);
        sprintf (result, "In-Reply-To: %s", msg_id);
        return result;
}



static char *
make_references (rstring_t *ref, char *msg_id)
{
        size_t   mi_len = 0;
        size_t   len    = 0;
        char   **iterator;
        char    *result;
        char    *result_ptr;

        if (msg_id)
                mi_len = strlen (msg_id);

        if (ref == NULL){
                result = xmalloc (12 + mi_len + 1);
                sprintf (result, "References: %s", msg_id);
                return result;
        }
  
        for (iterator = ref->array; *iterator; iterator++){
                len += 1 + strlen (*iterator);
        }

        result     = xmalloc (12 + len + 1 + mi_len + 1);
        result_ptr = stpcpy (result, "References:");
        for (iterator = ref->array; *iterator; iterator++){
                *result_ptr = ' ';
                result_ptr++;
                result_ptr  = stpcpy (result_ptr, *iterator);
        }
        *result_ptr = ' ';
        result_ptr++;
        result_ptr  = stpcpy (result_ptr, msg_id);
        *result_ptr = '\0';
        return result;
}


static char *
make_from (void)
{
        char *result;
        char *name   = ask_get_field (sender_ask, "my_name");
        char *email  = ask_get_field (sender_ask, "email");
        size_t n_len;
        size_t e_len;

        n_len = (name)  ? strlen (name)  : 0;
        e_len = (email) ? strlen (email) : 0;

        /*              From:_  name    _   <   email   >  \0 */
        result = xmalloc (6 +   n_len + 1 + 1 + e_len + 1 + 1);
  
        if (name && email){
                sprintf (result, "From: %s <%s>", name, email);
        }
        else if (email){
                sprintf (result, "From: %s", email);
        }
        else {
                sprintf (result, "From: ");
        }
        return result;
}



static char *
make_to (char *to)
{
        size_t  to_len = (to) ? strlen (to) : 0;
        char   *result = xmalloc (4 + to_len + 1);

        if (to)
                sprintf (result, "To: %s", to);
        else
                sprintf (result, "To: ");
        return result;
}



static int
may_be (address_t *addr, address_t *first, char *email)
{
        if (addr == NULL)
                return 0;

        if (addr->full == NULL)
                return 0;

        if (strcmp (addr->email, email) == 0)
                return 0;

        if (strcmp (addr->email, first->email) == 0)
                return 0;

        return 1;
}



static void
make_to_reply (rstring_t *strings, mail_t *mail, int all)
{
        int         i;
        char       *email = ask_get_field (sender_ask, "email");
        address_t  *first = NULL;
        raddress_t *to    = NULL;
        raddress_t *cc    = NULL;

        if (mail->to)
                to = raddress_create_size (1 + mail->to->count + 1);
        else
                to = raddress_create_size (2);
  
        if (mail->reply_to)
                first = mail->reply_to;
        else
                first = mail->from;

        if (first)
                raddress_add (to, first);

        if (all){
                for (i = 0; i < mail->to->count; i++){
                        if (may_be (mail->to->array[i], first, email))
                                raddress_add (to, mail->to->array[i]);
                }
                if (mail->cc && mail->cc->count > 0){
                        cc = raddress_create_size (mail->cc->count + 1);
                        for (i = 0; i < mail->cc->count; i++)
                                if (may_be (mail->cc->array[i], first, email))
                                        raddress_add (cc, mail->cc->array[i]);
                }
        }

        rstring_add (strings, raddress_list (to, "To", 1));
        raddress_destroy (to);
        if (cc){
                rstring_add (strings, raddress_list (cc, "Cc", 1));
                raddress_destroy (cc);
        }
}



static char *
make_subject (char *subj, char *prefix)
{
        size_t  s_len = 0;
        char   *result;
        char   *seek;

        if (subj)
                s_len = strlen (subj);

        result = xmalloc (14 + s_len + 1);
  
        if (subj && *subj){
                if (prefix)
                        seek = strstr (subj, prefix);
                else
                        seek = subj;

                if (! seek)
                        sprintf (result, "Subject: %s %s", prefix, subj);
                else
                        sprintf (result, "Subject: %s", subj);
        }
        else if (prefix) {
                sprintf (result, "Subject: %s", prefix);
        }
        else {
                sprintf (result, "Subject: ");
        }
        return result;
}




static char *
make_empty_line (void)
{
        return xstrdup ("");
}




static char *
make_wrote (mail_t *mail)
{
        char *result;
        char *fmt = NULL;

        fmt = address_wrote_format (mail->from);
        if (fmt == NULL)
                fmt = DEFAULT_REPLY_FMT;

        result = eprintf_mail (fmt, mail);
        return result;
}



static FILE *
open_file_or_pipe (const char *file)
{
        if (*file == '|')
                return popen (file + 1, "r");
        else
                return fopen (file, "r");
}



static void
close_file_or_pipe (const char *file, FILE *fp)
{
        if (*file == '|')
                pclose (fp);
        else
                fclose (fp);
}



static char *
make_signature (void)
{
        char *sigfile = ask_for_default ("sigfile", NULL);
        char *sig_dashes = ask_for_default ("sig_dashes", NULL);
        char *signature;
        int   ret;
        int   BUFSIZE = 1000; //Fundacja promowania dugich sygnaturek
        FILE *fp;

        if (sigfile && strcmp (sigfile, "default") == 0){
                signature = xstrdup ("\n-- \nElmo: A MUA that sucks much less.");
        }
        else if (!sigfile || sigfile == NULL){
                signature = NULL;
        }
        else {
                fp = open_file_or_pipe (sigfile);
                if (!fp){
                        error_ (errno, "%s", sigfile);
                        return NULL;
                }
                signature          = xmalloc (BUFSIZE);
                if (sig_dashes) {
                        strncpy (signature, "\n-- \n", 5);
                        ret                = fread (signature + 5, 1, BUFSIZE - 7, fp);
                        signature[ret + 5] = '\0';
                } else {
                        signature[0]       = '\n';
                        ret                = fread (signature + 1, 1, BUFSIZE - 2, fp);
                        signature[ret + 1] = '\0';
                }    
                close_file_or_pipe (sigfile, fp);
        }
        return signature;
}

static void
editor (char *fname, int lineno)
{
        char *command;
        char *editor = ask_for_default ("editor", NULL);
        char *env_editor = NULL;
        char *get_editor = NULL;
  
        if (!editor || !*editor){
                env_editor = getenv("EDITOR");
    
                if (env_editor != NULL){
                        editor = strncat(env_editor, " %f", 3);
                }
                else{
                        get_editor = xstrdup(read_argument ("editor: ", NULL, COMPLETE_FILES));
      
                        if (get_editor != NULL)
                                editor = strncat(get_editor, " %f", 3);
                        else
                                return;
                }
        }

        command = eprintf_editor (editor, fname, lineno);
        curs_set (1);
        endwin ();
  
        system (command);
        xfree (command);
  
        curs_set (0);
        folder_refresh ();
}



static void
write_to_file (FILE *fp, rstring_t *header, int body_start, int body_end,
               mail_t *reply)
{
        char *indent_string = ask_for_default ("indent_string", NULL);
        int   i;

        if (indent_string && reply)
                indent_string = eprintf_mail (indent_string, reply);

        for (i = 0; i < header->count; i++){

                if (i >= body_start && i <= body_end)
                        if (!indent_string) 
                                fprintf (fp, "> %s\n", header->array[i]);
                        else
                                fprintf (fp, "%s%s\n", indent_string, header->array[i]);
      
                else
                        fprintf (fp, "%s\n", header->array[i]);
        }

        if (indent_string && reply)
                xfree (indent_string);
}



static char *
make_file_and_run (rstring_t *header, int body_start, int body_end,
                   mail_t *reply)
{
        char *name;
        FILE *fp;

        fp = file_temp_file (& name);

        if (fp == NULL){
                rstring_free_strings (header);
                rstring_delete (header);
                return NULL;
        }
        write_to_file (fp, header, body_start, body_end, reply);
        fclose (fp);

        editor (name, body_start);
        return name;
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

char *
compose_new (char *to)
{
        char      *real_to;
        char      *ret;
        char      *subject;
        rstring_t *header;

        if (to == NULL)
                real_to = xstrdup (read_argument ("To: ", NULL, COMPLETE_ADDRS));
        else
                real_to = xstrdup (to);

        if (real_to == NULL)
                return NULL;
  
        subject = read_argument ("Subject: ", NULL, COMPLETE_NONE);
        if (subject == NULL){
                xfree (real_to);
                return NULL;
        }

        header      = rstring_create ();
        compose_act = COMPOSE_NEW;
  
        rstring_add (header, make_from ());
        rstring_add (header, make_to (real_to));
        rstring_add (header, make_subject (subject, NULL));
        rstring_add (header, make_empty_line ());
        rstring_add (header, make_signature ());

        ret = make_file_and_run (header, header->count, header->count, NULL);

        xfree (real_to);
        rstring_free_strings (header);
        rstring_delete (header);

        return ret;
}




char *
compose_reply (int all)
{
        mail_t    *mail = folder_mail_selected ();
        rstring_t *header;
        rstring_t *body;
        char      *signature;
        str_t     *mail_body;
        char      *ret;
        int        line_count = 4;
  
        PREAMBLE;
  
        header = rstring_create ();
  
        compose_act = COMPOSE_REPLY;
  
        if (mail->recv_for && mail->recv_for->email)
                ask_change_where (sender_ask, "email", mail->recv_for->email);
  
        rstring_add (header, make_from ());
        make_to_reply (header, mail, all);
        rstring_add (header, make_subject (mail->subject, "Re:"));
        line_count++;

        if (mail->msg_id){
                rstring_add (header, make_references (mail->in_reply_to, mail->msg_id));
                rstring_add (header, make_in_reply_to (mail->msg_id));
                line_count += 2;
        }
        rstring_add (header, make_empty_line ());
        rstring_add (header, make_wrote (mail));

        mail_body = wrapbox_mail_body (mail, -1);
        if (mail_body == NULL){
                error_ (errno, "%s", mail->place.file_name);
                rstring_delete (header);
                return NULL;
        }

        body      = rstring_split_lines (mail_body->str, 85);
        header    = rstring_join (header, body);
        signature = make_signature ();
        rstring_add (header, signature);
        ret       = make_file_and_run (header, line_count, header->count - 1,
                                       mail);

        xfree (signature);
        rstring_free_firstn_strings (header, line_count);
        rstring_delete (header);
  
        str_destroy (mail_body);
        return ret;
}




char *
compose_fwd (char *to)
{
        mail_t    *mail = folder_mail_selected ();
        rstring_t *header;
        rstring_t *body;
        str_t     *mail_body;
        char      *signature;
        char      *ret;
        int        line_count;

        PREAMBLE;
  
        if (to == NULL)
                to = read_argument ("To: ", NULL, COMPLETE_ADDRS);
        if (to == NULL)
                return NULL;
  
        header = rstring_create ();

        compose_act = COMPOSE_FWD;

        if (mail->recv_for && mail->recv_for->email)
                ask_change_where (sender_ask, "email", mail->recv_for->email);
  
        rstring_add (header, make_from ());
        rstring_add (header, make_to (to));
        rstring_add (header, make_subject (mail->subject, "Fwd:"));
        rstring_add (header, make_empty_line ());

        line_count = 4;
        mail_body  = wrapbox_mail_body (mail, -1);

        if (mail_body == NULL){
                error_ (errno, "%s", mail->place.file_name);
                rstring_delete (header);
                return NULL;
        }

        body      = rstring_split_lines (mail_body->str, 85);
        header    = rstring_join (header, body);
        signature = make_signature ();
        rstring_add (header, signature);
        ret       = make_file_and_run (header, 1, 0, mail);

        xfree (signature);
        rstring_free_firstn_strings (header, line_count);
        rstring_delete (header);

        str_destroy (mail_body);
        return ret;
}




char *
compose_date (void)
{
        return make_date ();
}




char *
compose_msg_id (void)
{
        return message_id ();
}



char *
compose_boundary (void)
{
        return boundary ();
}



void
compose_after_send (void)
{
        switch (compose_act){

                case COMPOSE_NEW:
                        break;

                case COMPOSE_REPLY:
                        folder_after_reply ();
                        break;

                case COMPOSE_FWD:
                        folder_after_fwd ();
                        break;

        }
}



/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE compose.c
 *
 ****************************************************************************/
