#!/usr/bin/env perl
#
#  elmoconf.pl - Configuration tool for Elmo.
#
#  Copyright (C) 2003-2004 Krzysztof Gibas <sdas % gibas org>
#                          rzyjontko
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 2.
#
#  This  program is  distributed in the  hope  that  it will be useful,
#  but  WITHOUT  ANY  WARRANTY;  without even  the  implied warranty of
#  MERCHANTABILITY  or  FITNESS  FOR  A  PARTICULAR  PURPOSE.  See  the
#  GNU General Public License for more details.
#
#  You  should  have  received a copy of the GNU General Public License
#  along  with  this  program;  if  not,  write  to  the  Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
#  USA.
#  
#  $Id: elmoconf.pl,v 1.24 2004/01/08 22:03:09 gibas Exp $ 
#  
#  --------------------------------------------------------------------
#
#  This script sets up user's config file based on few simple questions.

use strict;
use warnings;

use constant VERSION => "1.21";

#set secure environment
$ENV{'PATH'} = '/bin:/usr/bin:/usr/local/bin:/usr/pkg/bin';
delete @ENV{'IFS', 'CDPATH', 'ENV', 'BASH_ENV'};

my ($root, $ok, $start, $trash, $drafts, $sent, $outbox, $spam, $localbox, 
    $bayes_file, $guess_email, $editor, $charset, $addressbook, $signature, 
    $smtp_auth_user, $smtp_auth_passwd, $smtp_server, $pop_server, $protect,
    $my_name, $email, $smtp_auth, $username, $password, $pop_info_dir, $theme);

# This function takes 2 arguments.  First is the question (without trailing
# space or ask sign).  Second is the default value.  User will be shown the
# question and asked for a response.  If no answer is supplied, the default
# one will be returned.
sub ask {
    my $question = shift @_;
    my $default  = shift @_;

    print "$question [$default]: ";
    my $response = <STDIN>;
    chop $response;
    if ($response eq "" or $response =~ /^(\s+|\t+)$/) {
	return $default;
    }
    return $response;
}

sub ask_passwd {
    my ($response_1, $response_2);
    my $question = shift @_;
        
    do {
        print "$question: ";
        
        system("stty -echo 2>/dev/null");
        chop($response_1 = <STDIN>);
        system("stty echo 2>/dev/null");
        print "\n";
        
        print "Re-enter password: ";

        system("stty -echo 2>/dev/null");
        chop($response_2 = <STDIN>);
        system("stty echo 2>/dev/null");   
        print "\n";
        
	if ($response_1 ne $response_2) {
	    print "They don't match; try again.\n";
	}
	
    } while ($response_1 ne $response_2);
    
    return $response_1;
}

# Takes only 1 argument and replaces leading tilde (~) with users home
# directory.
sub replace_tilde {
    $_ = shift @_;

    s/^~/$ENV{HOME}/g;
    return $_;
}

sub have_space {
    $_ = shift @_;

    if (/.*\s+.*/ || length($_) == 0) {
        return "\'$_\'";
    }
    else {
        return $_;
    }
}

sub is_valid_maildir {
    my $dir = shift @_;

    return -d $dir && -d "$dir/cur" && -d "$dir/new" && -d "$dir/tmp";
}


sub make_maildir {
    my $dir = shift @_;

    if (-f $dir) {
        unlink ($dir) || die "\nCouldn't delete file: $dir\n";
    }
    if (-d $dir) {
	system ('rm', '-rf', "$dir") || die "\nCouldn't delete directory: $dir\n";
    }
    
    mkdir ("$dir", 0700) || die "\nCouldn't create $dir: $!\n";
    mkdir ("$dir/new", 0700) || die "\nCouldn't create $dir/new: $!\n";
    mkdir ("$dir/cur", 0700) || die "\nCouldn't create $dir/cur: $!\n";
    mkdir ("$dir/tmp", 0700) || die "\nCouldn't create $dir/tmp: $!\n";
}


sub print_valid {
    my $dir = shift @_;

    print "\nThe directory $dir seems to be a valid maildir.\n";
    print "Using it as a $dir directory.\n\n";
}

sub print_dir_exists {
    my $dir = shift @_;

    print "\nThere is a $dir directory under the specified directory,\n";
    print "but it is not a valid maildir. Maybe another program is using\n";
    print "this directory to store your messages, but using different\n";
    print "format. You will be asked for a box. If $dir is your answer,\n";
    print "it will be destroyed.\n\n"
}

sub print_file_exists {
    my $file = shift @_;

    print "\nThere is a file $file under the specified directory.\n";
    print "You can't use it as a $file because elmo won't be able to\n";
    print "move messages there. You will be asked for a box. If $file\n";
    print "is your answer, it will be removed.\n\n";
}

sub print_only_name {
    print "\nPlease specify only a name of the box under the specified\n";
    print "directory. If it doesn't exist, it will be created for\n";
    print "you. This name must be relative and cannot contain / sign.\n\n";
}


sub local_box {
    my $localbox = shift @_;

    print "\nIt seems that you have a local spool mailbox located at\n";
    print "$localbox. You can read mail delivered to $localbox\n";
    print "with elmo, but there must be a symbolic link created.\n\n";
}


sub print_help {
    print "\nelmoconf.pl - Configuration tool for Elmo.\n\n";
    print "Usage: elmoconf.pl [option]\n\n";
    print "Options:\n";
    print "\t-g | --global\t\tgenerate global config in /etc/elmorc\n";
    print "\t-h | --help\t\tprints this help\n";
    print "\t-p | --pop\t\tadd new pop3 account to config\n";
    print "\t-s | --smtp\t\tadd new smtp account to config\n";
    print "\t-V | --version\t\tshows version\n\n";
    exit 0;
}

sub print_version {
    print "elmoconf.pl-" . VERSION . "\n";
    exit 0;
}

sub get_smtp_info {
    if ($ENV{EMAIL}){
        $guess_email = $ENV{EMAIL};
    }
    elsif ($ENV{USER} && $ENV{HOSTNAME}){
        $guess_email = "$ENV{USER}" . "\@" . "$ENV{HOSTNAME}";
    }
    elsif ($ENV{USER} && $ENV{HOST}){
        $guess_email = "$ENV{USER}" . "\@" . "$ENV{HOST}";
    }
    else {
        $guess_email = "";
    }

    $my_name     = ask ("What is your name", "$ENV{USER}");
    $email       = ask ("What is your email", "$guess_email");

    if ($ENV{HOSTNAME}) {
        $smtp_server = ask ("What is your smtp server", "$ENV{HOSTNAME}");
    }
    elsif ($ENV{HOST} && !$ENV{HOSTNAME}) {
        $smtp_server = ask ("What is your smtp server", "$ENV{HOST}");
    }
    else{
        $smtp_server = ask ("What is your smtp server", "");
    }

    $smtp_auth   = ask ("Should I use smtp auth for your account", "yes");

    if ($smtp_auth eq "yes") {
        $smtp_auth_user  = ask ("What is your smtp auth username", "$ENV{USER}");
        $smtp_auth_passwd = ask_passwd ("What is your smtp auth password");
    }
    else {
        $smtp_auth_user = $ENV{USER};
        $smtp_auth_passwd = "";
    }
}

sub write_smtp_info {
    print CONFIG "set smtp_acc {\n  name: $smtp_server\n  server: $smtp_server\n";
    print CONFIG "  my_name: " . have_space($my_name) . "\n  email: $email\n  port: 25\n";

    if ($smtp_auth eq "yes") {
        print CONFIG "  username: " . have_space($smtp_auth_user) . "\n";
        print CONFIG "  password: " . have_space($smtp_auth_passwd) . "\n";
    }

    print CONFIG "\}\n\n";    
}

sub get_pop_info {
    if ($ENV{HOSTNAME}) {
        $pop_server = ask ("What is your pop3 server", "$ENV{HOSTNAME}");
    }
    elsif ($ENV{HOST}) {
        $pop_server = ask ("What is your pop3 server", "$ENV{HOST}");
    }
    else{
        $pop_server = ask ("What is your pop3 server", "");
    }

    $username   = ask ("What is your pop3 username", "$ENV{USER}");
    $password   = ask_passwd ("What is your pop3 password");        
}

sub write_pop_info {
    print CONFIG "set pop_acc {\n  name: $pop_server\n  server: $pop_server\n";
    print CONFIG "  username: " . have_space($username) . "\n  password: " . have_space($password) . "\n  port: 110\n";
    print CONFIG "  count_max: 0\n}\n\n";
}

sub write_mailbox {
    print CONFIG "set mailbox {\n  name: 'mail'\n  root: " . have_space($root) . "\n";
    print CONFIG "  start: " . have_space($start) . "\n  drafts: " . have_space($drafts) . "\n";
    print CONFIG "  trash: " . have_space($trash) . "\n  sent: " . have_space($sent) . "\n";
    print CONFIG "  outbox: " . have_space($outbox) . "\n  spam: " . have_space($spam) . "\n";
    print CONFIG "  protect: " . have_space($protect) . "\n}\n\n";
}

sub write_sigfile {
    print CONFIG "set sigfile " . have_space($signature) . "\n";
}

sub write_editor {
    print CONFIG "set editor \'$editor \%f\'\n";
}

sub write_addressbook {
    print CONFIG "set addressbook " . have_space($addressbook) . "\n";
}

sub write_pop_infodir {
    print CONFIG "set pop_info_dir " . have_space($pop_info_dir) . "\n\n";
}

sub write_bayes_file {
    if ($bayes_file and $protect eq "yes") {
        print CONFIG "set bayes_file " . have_space($bayes_file) . "\n";
    }
}

sub write_other {
    print CONFIG "set indent_string '> '\n";
    print CONFIG "set sig_dashes on\n";
    print CONFIG "#set welcome_msg no\n";
    print CONFIG "set relative_names on\n\n";
}

sub write_formats {
    print CONFIG "#set line_format \'\%\?\%\$\%\# \%D \%016f \(\%\-06S\)  \'\n";
    print CONFIG "#set sent_format \'\%\?\%\$  \%D \%016t \(\%-06S\)  \'\n";
    print CONFIG "#set abook_format \'\%s \%020n \%e\'\n";
    print CONFIG "#set win_fetch {\n\#\tfetch_fmt: \'\%\?\%\$  \%D \%016f \(\%-06S\)  \%s\'\n\#}\n";
    print CONFIG "#set win_help {\n\#\thelp_fmt: \'\%012k \%030f \%d\'\n\#}\n";
    print CONFIG "#set win_boxes {\n\#\tbox_fmt: \' \%010n \(\%u\/\%t\)\'\n\#}\n";
    print CONFIG "#set win_attach {\n\#\tattach_fmt: \'\%020f \(\%-06s\) [\%t\%C\%E]\'\n\#}\n";
    print CONFIG "#set win_sender {\n\#\tsender_fmt: \'\%020f \(\%-06s\) [\%t\%C\%E]\'\n\#}\n\n";
}


sub get_charset {
    my %LANG = (
                "ca_ES" => "iso-8859-1",
                "cs_CZ" => "iso-8859-2",
                "da_DK" => "iso-8859-1",
                "de_DE" => "iso-8859-1",
                "el_GR" => "iso-8859-7",
                "es_ES" => "iso-8859-1",
                "et_EE" => "iso-8859-1",
                "fi_FI" => "iso-8859-1",
                "fr_FR" => "iso-8859-1",
                "gl_ES" => "iso-8859-1",
                "he_IL" => "iso-8859-8",
                "hr_HR" => "iso-8859-2",
                "hu_HU" => "iso-8859-2",
                "is_IS" => "iso-8859-1",
                "it_IT" => "iso-8859-1",
                "lt_LT" => "iso-8859-13",
                "nl_NL" => "iso-8859-1",
                "nn_NO" => "iso-8859-1",
                "no_NO" => "iso-8859-1",
                "pl_PL" => "iso-8859-2",
                "pt_PT" => "iso-8859-1",
                "ro_RO" => "iso-8859-2",
                "ru_RU" => "iso-8859-5",
                "sk_SK" => "iso-8859-2",
                "sl_SI" => "iso-8859-2",
                "sv_SE" => "iso-8859-1",
                "tr_TR" => "iso-8859-9"
        );

    if (($ENV{LANG} && $ENV{LANG} eq 'C') || ($ENV{LC_ALL} && $ENV{LC_ALL} eq 'C') || ($ENV{LC_CTYPE} && $ENV{LC_CTYPE} eq 'C')) {
        $charset = "US-ASCII";
    }
    elsif ($ENV{LC_ALL} and exists ($LANG{"$ENV{LC_ALL}"})) {
        $charset = $LANG{"$ENV{LC_ALL}"};
    }
    elsif ($ENV{LC_CTYPE} and exists ($LANG{"$ENV{LC_CTYPE}"})) {
        $charset = $LANG{"$ENV{LC_CTYPE}"};
    }
    elsif ($ENV{LANG} and exists ($LANG{"$ENV{LANG}"})) {
        $charset = $LANG{"$ENV{LANG}"};
    }
}

sub write_charset {
    print CONFIG "set charset " . have_space($charset) . "\n";
    print CONFIG "translate utf-8 " . have_space($charset) . "\n";

    if ($charset eq "iso-8859-2" || $charset eq "ISO-8859-2") {
        print CONFIG "translate windows-1250 iso-8859-2\n";
    }
}

sub get_theme {
    print "Themes:\n";
    print "\t80x25 - simple, created for 80x25 consoles.\n";
    print "\toutlook - looks nice on framebuffer console.\n";
    $theme = ask ("What theme do you choose", "default");
}

sub write_theme {
my $theme_80x25 = "set line_format \"\%\$ \%D \%014f (\%-06S) \"\nset sent_format \"\%\$ \%D \%014f (\%-06S) \"\nset win_frames \{\n
\tfill:\"(15 0 15 0)\"\n\tf_horiz: \"(0 -6 14 -6)\"\n\tf_vert: \"(15 1 15 -1)\"\n\tf_left: \"(15 -6 15 -6)\"\n}\n\nset win_clock \{\n
\twidth: 15\n\theight: 5\n\ttop: -7\n\tclock_fmt: \" \%H:\%M \%e \%b \"\n}\n\nset win_folder \{\n\twidth: -16\n\tleft: 16\n}\n\n
set win_boxes \{\n\twidth: 15\n\theight: 14\n\tbox_fmt: \" \%010n\"\n}\n\nset win_fetch \{\n\tfetch_fmt: \"\%\$ \%D \%014f (\%-06S) \"\n}";

my $theme_outlook = "set win_folder \{\n\theight: 15\n}\n\nset win_mail \{\n\tleft: 26\n\ttop: 18\n\theight: -20\n\twidth: -26\n}\n\n
set win_boxes \{\n\theight: 15\n}\n\nset win_frames \{\n\tf_horiz: \"(0 16 -1 16) (0 -8 24 -8)\"\n\tf_left: \"(25 -8 25 -8)\"\n
\tf_cross: \"(25 16 25 16)\"\n}\n\nset win_attach \{\n\tleft: 26\n\twidth: -26\n\}\n\nkey mail q cmd_quit";

    if ($theme ne "default") {
        if (! -e "$root/.themes") {
            mkdir("$root/.themes", 0700) || die "\nCouldn't create $root/.themes: $!\n"
        }
        
        if (! -e "$root/.themes/80x25" && $theme eq "80x25") {
            open (THEME, ">$root/.themes/80x25") || die "\nCouldn't create theme 80x25: $!\n";
            flock (THEME, 2) || die "\nCouldn't lock file theme 80x25: $!";
            
                print THEME $theme_80x25;
            
            close (THEME) || print "\nCouldn't close theme file: $!\n";
        }
        if (! -e "$root/.themes/outlook" && $theme eq "outlook") {
            open (THEME, ">$root/.themes/outlook") || die "\nCouldn't create theme outlook: $!\n";
            flock (THEME, 2) || die "\nCouldn't lock file theme outlook: $!";

                print THEME $theme_outlook;

            close (THEME) || print "\nCouldn't close theme file: $!\n";
        }

        print CONFIG "\n#theme\n";
        print CONFIG "include $root/.themes/$theme\n\n";
    }
}

sub write_key_bindings {
    print CONFIG "#KEY BINDINGS\n";
    print CONFIG "key mail r sender_open_reply\n";
    print CONFIG "key mail f sender_open_fwd\n";
    print CONFIG "key mail R sender_open_reply_all\n\n";
}

sub write_hooks {
    print CONFIG "#HOOKS\n";
    print CONFIG "hook folder_toggle_flag folder_bar_next\n";
    print CONFIG "hook fetch_del_mail fetch_next\n";
    print CONFIG "hook fetch_get_mail fetch_next\n";
    print CONFIG "hook mybox_create_box box_selection_re_read\n";
    print CONFIG "hook sender_go mailreader_close\n";
    print CONFIG "hook abook_hit abook_next\n\n";
}

sub check_config_exists {
    my $dir = shift @_;
    my $file = shift @_;
    
    if (-e "$dir/$file") {
        print "\nFound existing $dir/$file. The existing file will be\n";
        print "renamed to $dir/$file.0.\n\n";
        rename "$dir/$file", "$dir/$file.0";
    }
}


# STEP 0
# PROGRAM PARAMETRS
if (@ARGV >= 1) {
    if ($ARGV[0] eq '--global' or $ARGV[0] eq '-g') {
        #generate global config in /etc/elmorc
        if ($< != 0) {
            print "You must be root in order to write in /etc directory.\n";
            exit 0;
        }

        check_config_exists("/etc", "elmorc");
        
        open (CONFIG, ">/etc/elmorc") || die "\nCouldn't create /etc/elmorc: $!\n";
        flock (CONFIG, 2) || die "\nCouldn't lock file /etc/elmorc: $!";
        
            $root = "~/mail";
            $start = "inbox";
            $drafts = "drafts";
            $trash = "trash";
            $sent = "sent";
            $outbox = "outbox";
            $spam = "spam";
            $protect = "yes";
            write_mailbox();
 
            $smtp_server = "localhost";
            $my_name = "\$USER";
            $email = "\$USER\@\$HOSTNAME";        
            $smtp_auth = "no";
            write_smtp_info();
            
            $bayes_file = "~/mail/.bayes_file";
            write_bayes_file();
        
            $addressbook = "~/mail/.addressbook";
            write_addressbook();
        
            $pop_info_dir = "~/mail/.pop";
            write_pop_infodir();

            $signature = "~/.signature";
            write_sigfile();

            $editor = "\$EDITOR";
            write_editor();
        
            write_other();

            # CHARSET
            get_charset();

            if (! $charset) {
                $charset = "iso-8859-1";
            }
            write_charset();
        
            # KEY BINDINGS
            write_key_bindings();

            # HOOKS
            write_hooks(); 
    
        close (CONFIG) || print "\nCouldn't close file: $!\n";
        chmod (0644, "/etc/elmorc") or die "\nCouldn't chmod /etc/elmorc: $!\n";        
        exit 0;
    }
    elsif ($ARGV[0] eq '--help' or $ARGV[0] eq '-h') {
        #shows help.
        print_help();
    }
    elsif ($ARGV[0] eq '--pop' or $ARGV[0] eq '-p') {
        #add new pop3 account.
        get_pop_info();
        
        open (CONFIG, ">>$ENV{HOME}/.elmorc") || die "\nCouldn't add to file $ENV{HOME}/.elmorc: $!\n";
        flock (CONFIG, 2) || die "\nCouldn't lock file $ENV{HOME}/.elmorc: $!";
        
        write_pop_info();
        
        close (CONFIG) || print "\nCouldn't close file: $!\n";
        chmod (0600, "$ENV{HOME}/.elmorc") or die "\nCouldn't chmod $ENV{HOME}/.elmorc: $!\n";
        
        exit 0;
    }
    elsif ($ARGV[0] eq '--smtp' or $ARGV[0] eq '-s') {
        #add new smtp account.
        get_smtp_info();

        open (CONFIG, ">>$ENV{HOME}/.elmorc") || die "\nCouldn't add to file $ENV{HOME}/.elmorc: $!\n";
        flock (CONFIG, 2) || die "\nCouldn't lock file $ENV{HOME}/.elmorc: $!";
        
        write_smtp_info();
        
        close (CONFIG) || print "\nCouldn't close file: $!\n";
        chmod (0600, "$ENV{HOME}/.elmorc") or die "\nCouldn't chmod $ENV{HOME}/.elmorc: $!\n";

        exit 0;
    }
    elsif ($ARGV[0] eq '--version' or $ARGV[0] eq '-V') {
        #show version
        print_version();
    }
    else {
        print "elmoconf.pl: unrecognized option \`" . $ARGV[0] . "\'\n";
        print "Try `elmoconf.pl --help' for more information.\n";
        exit 0;
    }
}
 
# STEP 1
# Set up root directory of mailbox.
do {
    $root = ask ("Where do you want to store your mailboxes", "~/mail");
    $root = replace_tilde ($root);

    if (-d $root) {
	$ok = 1;
    }
    elsif (! -e $root) {
	mkdir ($root, 0700) || die "\nCouldn't create $root: $!\n";
	$ok = 1;
    }
    else {
	print "\nYou have specified an existing file to be a location\n";
	print "of your mailboxes collection. I'm afraid it's impossible.\n";
	print "Elmo stores your mailboxes in a single directory.\n\n";
	print "Please specify a directory name.\n\n"
    }
} while (! $ok);


# STEP 2
# Start box.
if (is_valid_maildir ("$root/inbox")) {
    print_valid ("inbox");
    $start = "inbox";
}
elsif (-d "$root/inbox") {
    print_dir_exists ("inbox");
    $start = ask ("What is your start box", "inbox");
    make_maildir ("$root/$start");
}
elsif (-f "$root/inbox") {
    print_file_exists ("inbox");
    $start = ask ("What is your start box", "inbox");
    make_maildir ("$root/$start");
}
else {
    $start = ask ("What is your start box", "inbox");
    while (index ($start, "/") != -1) {
	print_only_name();
	$start = ask ("What is your start box", "inbox");
    }
    make_maildir ("$root/$start");
}

# STEP 3
# Trash.
if (is_valid_maildir ("$root/trash")) {
    print_valid ("trash");    
    $trash = "trash";
}
elsif (-d "$root/trash") {
    print_dir_exists ("trash");
    $trash = ask ("Where do you want to move deleted messages", "trash");
    make_maildir ("$root/$trash");
}
elsif (-f "$root/trash") {
    print_file_exists ("trash");
    $trash = ask ("Where do you want to move deleted messages", "trash");
    make_maildir ("$root/$trash");
}
else {
    $trash = ask ("Where do you want to move deleted messages", "trash");
    while (index ($trash, "/") != -1) {
	print_only_name();
	$trash = ask ("Where do you want to move deleted messages", "trash");
    }
    make_maildir ("$root/$trash");
}

# STEP 4
# Drafts.
if (is_valid_maildir ("$root/drafts")) {
    print_valid ("drafts");
    $drafts = "drafts";
}
elsif (-d "$root/drafts") {
    print_dir_exists ("drafts");
    $drafts = ask ("Where do you want to store your unfinished messages", "drafts");
    make_maildir ("$root/$drafts");
}
elsif (-f "$root/drafts") {
    print_file_exists ("drafts");
    $drafts = ask ("Where do you want to store your unfinished messages", "drafts");
    make_maildir ("$root/$drafts");
}
else {
    $drafts = ask ("Where do you want to store your unfinished messages", "drafts");
    while (index ($drafts, "/") != -1) {
	print_only_name();
	$drafts = ask ("Where do you want to store your unfinished messages", "drafts");
    }
    make_maildir ("$root/$drafts");
}

# STEP 5
# Sent.
if (is_valid_maildir ("$root/sent")) {
    print_valid ("sent");
    $sent = "sent";
}
elsif (-d "$root/sent") {
    print_dir_exists ("sent");
    $sent = ask ("Where do you want to store your sent messages", "sent");
    make_maildir ("$root/$sent");
}
elsif (-f "$root/sent") {
    print_file_exists ("sent");
    $sent = ask ("Where do you want to store your sent messages", "sent");
    make_maildir ("$root/$sent");
}
else {
    $sent = ask ("Where do you want to store your sent messages", "sent");
    while (index ($sent, "/") != -1) {
	print_only_name();
    	$sent = ask ("Where do you want to store your sent messages", "sent");
    }
    make_maildir ("$root/$sent");
}



# STEP 6
# Outbox
if (is_valid_maildir ("$root/outbox")) {
    print_valid ("outbox");
    $outbox = "outbox";
}
elsif (-d "$root/outbox") {
    print_dir_exists ("outbox");
    $outbox = ask ("Where do you want to store your outbox messages", "outbox");
    make_maildir ("$root/$outbox");
}
elsif (-f "$root/outbox") {
    print_file_exists ("outbox");
    $outbox = ask ("Where do you want to store your outbox messages", "outbox");
    make_maildir ("$root/$outbox");
}
else {
    $outbox = ask ("Where do you want to store your outbox messages", "outbox");
    
    while (index ($outbox, "/") != -1) {
        print_only_name();
        $outbox = ask ("Where do you want to store your outbox messages", "outbox");
    }
    make_maildir ("$root/$outbox");
}

# STEP 7
# Spam.
if (is_valid_maildir ("$root/spam")) {
    print_valid ("spam");
    $spam = "spam";
}
elsif (-d "$root/spam") {
    print_dir_exists ("spam");
    $spam = ask ("Where do you want to store your spam messages", "spam");
    make_maildir ("$root/$spam");
}
elsif (-f "$root/spam") {
    print_file_exists ("spam");
    $spam = ask ("Where do you want to store your spam messages", "spam");
    make_maildir ("$root/$spam");
}
else {
    $spam = ask ("Where do you want to store your spam messages", "spam");
    while (index ($spam, "/") != -1) {
	print_only_name();
	$spam = ask ("Where do you want to store your sent messages", "sent");
    }
    make_maildir ("$root/$spam");
}

# STEP 8
# Existing spool mailboxes
if ($ENV{MAILDIR} && -e $ENV{MAILDIR}) {
    local_box ($ENV{MAILDIR});
    $localbox = ask ("How do you want to name your local box", "local");
    symlink ("$ENV{MAILDIR}", "$root/$localbox");
}
if ($ENV{MAIL} && -e $ENV{MAIL}) {
    local_box ($ENV{MAIL});
    $localbox = ask ("How do you want to name your local box", "local");
    symlink ("$ENV{MAIL}", "$root/$localbox");
}


# STEP 9
# SMTP
get_smtp_info();


# STEP 10
# POP3
get_pop_info();


# STEP 11
# SPAM PROTECTION
$protect     = ask ("Should I protect your mail against spam", "yes");

if ($protect eq "yes") {
    $bayes_file  = "$root/.bayes_file";
}

# STEP 12
# ADDRESSBOOK
$addressbook  = ask ("Where do you want store your addressbook", "$root/.addressbook");
$addressbook = replace_tilde($addressbook);

if (! -e $addressbook) {
    open(ADDRESSBOOK, ">$addressbook") || print "\nCouldn't create addressbook: $!\n";
    close(ADDRESSBOOK) || print "\nCouldn't close $addressbook: $!\n";
}

# STEP 13
# EDITOR
if ($ENV{EDITOR}) {
  $editor  = ask ("What is your favorite editor", "$ENV{EDITOR}");
}
else {
  $editor  = ask ("What is your favorite editor", "vim");
}
  
# STEP 14
# SIGNATURE
$signature = ask ("Where is your signature file", "$ENV{HOME}/.signature");
$signature = replace_tilde($signature);

# STEP 15
# CHARSET
get_charset();
if (! $charset) {
    $charset = ask ("What is your charset", "iso-8859-1");
}

# STEP 15
# THEMES
get_theme();

# STEP 16
# POP_INFO_DIR
if (! -d "$root/.pop") {
    mkdir ("$root/.pop", 0700) || print "\nCouldn't create pop_info_dir - $root/.pop: $!\n";
}
$pop_info_dir = "$root/.pop";


# STEP 17
# WRITE TO CONFIG
check_config_exists($ENV{HOME}, ".elmorc");

open (CONFIG, ">$ENV{HOME}/.elmorc") || die "\nCouldn't create $ENV{HOME}/.elmorc: $!\n";
flock (CONFIG, 2) || die "\nCouldn't lock file $ENV{HOME}/.elmorc: $!";

    write_mailbox();

    write_smtp_info();

    write_pop_info();

    write_pop_infodir();

    write_bayes_file();

    write_addressbook();

    write_sigfile();

    write_editor();

    write_charset();

    write_theme();

    write_other();

    # FORMATS
    write_formats();

    # KEY BINDINGS
    write_key_bindings();

    # HOOKS
    write_hooks();

close (CONFIG) || print "\nCouldn't close file: $!\n";

chmod (0600, "$ENV{HOME}/.elmorc") or die "\nCouldn't chmod $ENV{HOME}/.elmorc: $!\n";

