/* 
   elmo - Electronic Mail Operator

   Copyright (C) 2003, 2004 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <errno.h>

#ifdef HAVE_ICONV
# include <iconv.h>
#endif

#include "mime.h"
#include "xmalloc.h"
#include "file.h"
#include "memchunk.h"
#include "ask.h"
#include "error.h"
#include "str.h"
#include "misc.h"
#include "gettext.h"
#include "debug.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define isupper(a)  ((a) >= 'A' && (a) <= 'Z')
#define islower(a)  ((a) >= 'a' && (a) <= 'z')
#define isdigit(a)  ((a) >= '0' && (a) <= '9')
#define isxdigit(a) (isdigit (a) || ((a) >= 'A' && (a) <= 'F') \
                     || ((a) >= 'a' && (a) <= 'f'))
#define isbase64(a) (isupper (a) || islower (a) || isdigit (a) || (a) == '+' \
                     || (a) == '/' || (a) == '=')

#define BASE64_VAL(a) (isupper (a) ? ((a) - 'A') \
                       : (islower (a) ? ((a) - 'a' + 26) \
                          : (isdigit (a) ? ((a) - '0' + 52) \
                             : ((a) == '+' ? 62 \
                                : ((a) == '/' ? 63 : 0)))))

#define BASE64_SIGN(a) (((a) < 26) ? ('A' + (a)) \
                        : (((a) < 52) ? ('a' + (a) - 26) \
                           : (((a) < 62) ? ('0' + (a) - 52) \
                              : (((a) == 62) ? '+' \
                                 : (((a) == 63) ? '/' : 0)))))
                        

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

struct pair_list {
        struct pair_list *next;
        char             *from;
        char             *to;
};


struct rfc2047 {
        int         special_;   /* 1 if _ means a space */
        char       *charset;    /* charset used in first occurence */
        char       *str;        /* header string */
        str_t      *inter;      /* intermediate buffer */
        str_t      *result;     /* resulting buffer */
        regmatch_t  matches[5]; /* 0 - all, 2 - charset,
                                   3 - Q/B, 4 - content */
};


struct mime_handler {
        struct mime_handler *next;
        regex_t              re;
        char                *type;
        char                *handler;
};

/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/* These are used to find occurences of something like: =?...?Q?...?= */
static int     encoded_compiled = 0;
static regex_t encoded_string_re;

/* These are used to find mime types like "text/plain; charset=iso-8859-2" */
static int     mime_type_compiled = 0;
static regex_t mime_type_re;

/* List of charset conversions */
static struct pair_list *conversions = NULL;

static struct pair_list *prepared = NULL;

/* List of MIME handlers. */
static struct mime_handler *handlers = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    QP / BASE64 FUNCTIONS (RFC 2045)
 ****************************************************************************/

/**
 * to -- inclusive
 */
static void
decode_qp (str_t *dest, char *from, char *to, int special_)
{
        char buf[3];

        buf[2] = '\0';
  
        while (*from && from <= to){

                switch (*from){

                        case '_':
                                if (special_){
                                        str_put_char (dest, ' ');
                                }
                                else {
                                        str_put_char (dest, '_');
                                }
                                from++;
                                break;
      
                        case '=':
                                if (isxdigit (from[1]) && isxdigit (from[2])){
                                        buf[0] = from[1];
                                        buf[1] = from[2];
                                        str_put_char (dest,
                                                      strtol (buf, NULL, 16));
                                        from += 3;
                                }
                                else if (from[1] == '\n'){
                                        from += 2;
                                }
                                else {
                                        from += 3;
                                }
                                break;

                        default:
                                str_put_char (dest, *from);
                                from++;
                                break;
                }
        }
}



static void
decode_base64 (str_t *dest, char *from, char *to)
{
        while (from <= to && ! isbase64 (*to))
                to--;
        
        while (*from && from <= to){
                while (*from == '\n' || *from == '\r')
                        from++;
                str_put_char (dest, (BASE64_VAL (from[0]) << 2)
                                  | (BASE64_VAL (from[1]) >> 4));

                if (from[2] == '=')
                        break;
                str_put_char (dest, (BASE64_VAL (from[1]) << 4)
                                  | (BASE64_VAL (from[2]) >> 2));

                if (from[3] == '=')
                        break;
                str_put_char (dest, (BASE64_VAL (from[2]) << 6)
                                  | (BASE64_VAL (from[3]) >> 0));
                from += 4;
        }
}



static void
encode_base64 (str_t *dest, unsigned char *from, unsigned char *to)
{
        int line_len = 0;

        while (from <= to){

                str_put_char (dest, BASE64_SIGN (0x3f & (from[0] >> 2)));

                if (from + 1 > to){
                        str_put_char (dest, BASE64_SIGN (0x3f & (from[0] << 4)));
                        str_put_nchar (dest, '=', 2);
                        break;
                }
                str_put_char (dest, BASE64_SIGN (0x3f & ((from[0] << 4)
                                                         | (from[1] >> 4))));

                if (from + 2 > to){
                        str_put_char (dest, BASE64_SIGN (0x3f & (from[1] << 2)));
                        str_put_char (dest, '=');
                        break;
                }
                str_put_char (dest, BASE64_SIGN (0x3f & ((from[1] << 2)
                                                         | (from[2] >> 6))));
                str_put_char (dest, BASE64_SIGN (0x3f &  (from[2])));

                if (line_len == 76){
                        str_put_char (dest, '\n');
                        line_len = 0;
                }
                else {
                        line_len += 4;
                }
                from += 3;
        }
}

/****************************************************************************
 *    LIST FUNCTIONS
 ****************************************************************************/

static void
list_destroy (struct pair_list *list)
{
        if (list == NULL)
                return;

        list_destroy (list->next);

        if (list->from)
                xfree (list->from);
        if (list->to)
                xfree (list->to);
        xfree (list);
}


static struct pair_list *
list_find (struct pair_list *list, char *from)
{
        if (list == NULL)
                return NULL;

        if (strcmp (list->from, from) == 0)
                return list;

        return list_find (list->next, from);
}


static void
destroy_handler (struct mime_handler *list)
{
        if (list == NULL)
                return;

        destroy_handler (list->next);

        if (list->type)
                xfree (list->type);
        if (list->handler)
                xfree (list->handler);
        regfree (& list->re);
        xfree (list);
}


static struct mime_handler *
find_type (struct mime_handler *list, const char *fname)
{
        int ret;
        
        if (list == NULL)
                return NULL;

        ret = regexec (& list->re, fname, 0, NULL, 0);

        if (ret && ret != REG_NOMATCH){
                error_regex (ret, & list->re, NULL);
                return NULL;
        }
        else if (ret)
                return find_type (list->next, fname);

        return list;
}



static struct mime_handler *
find_mime (struct mime_handler *list, const char *type)
{
        if (list == NULL)
                return NULL;

        if (strcmp (list->type, type) == 0)
                return list;

        return find_mime (list->next, type);
}

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


static int
find_encoded_string (struct rfc2047 *x)
{
        int ret;

        if (! encoded_compiled)
                return 0;

        ret = regexec (& encoded_string_re, x->str, 5, x->matches, 0);

        if (ret && ret != REG_NOMATCH){
                error_regex (ret, & encoded_string_re, NULL);
                return 0;
        }

        return ! ret;
}



static void
save_charset (struct rfc2047 *s)
{
        int len = s->matches[2].rm_eo - s->matches[2].rm_so;
        
        s->charset = xmalloc (len + 1);

        memcpy (s->charset, s->str + s->matches[2].rm_so, len);
        s->charset[len] = '\0';
}



static void
get_intermediate (struct rfc2047 *s)
{
        char *from;
        char *to;
        
        if (s->inter)
                str_clear (s->inter);
        else
                s->inter = str_create ();

        str_put_string_len (s->inter, s->str, s->matches[0].rm_so);

        while (1) {

                from = s->str + s->matches[4].rm_so;
                to   = s->str + s->matches[4].rm_eo - 1;
                
                switch (s->str[s->matches[3].rm_so]){

                        case 'q':
                        case 'Q':
                                decode_qp (s->inter, from, to, s->special_);
                                break;

                        case 'b':
                        case 'B':
                                decode_base64 (s->inter, from, to);
                                break;
                }

                s->str += s->matches[0].rm_eo;
                if (! find_encoded_string (s))
                        break;

                str_put_string_len (s->inter, s->str, s->matches[0].rm_so);
        }

        str_put_string (s->inter, s->str);
}


#ifdef HAVE_ICONV
static void
translate (struct rfc2047 *s, char *to)
{
        int      ret;
        iconv_t  id = iconv_open (to, s->charset);
        char    *inptr, *outptr, *result;
        int      insize, outsize, ressize;
        int      bad_chars = 0;

        if (id == (iconv_t) -1){
                if (errno == EINVAL)
                        error_ (0,
                                _("conversion from %s to %s not available"),
                                s->charset, to);
                else
                        error_ (errno, _("couldn't convert from %s to %s"),
                                s->charset, to);
                return;
        }

        inptr   = s->inter->str;
        insize  = s->inter->len;

        result  = xmalloc (insize + 1);
        outptr  = result;
        outsize = ressize = insize;


        while (insize > 0){
                ret = iconv (id, & inptr, & insize, & outptr, & outsize);
                if (ret == (size_t) -1 && errno != E2BIG){
                        bad_chars++;
                        inptr++;
                        insize--;
                }
                if (errno == E2BIG){
                        result   = xrealloc (result, ressize * 2);
                        outptr  += ressize - outsize;
                        outsize += ressize;
                        ressize *= 2;
                }
        }

        if (bad_chars){
                error_ (0, _("some characters were not converted from %s "
                             "to %s due to invalid multibyte sequence"),
                        s->charset, to);
        }

        *outptr   = '\0';
        s->result = str_dup (result);

        xfree (result);
        iconv_close (id);
}
#endif



static void
get_result (struct rfc2047 *s)
{
        struct pair_list *list = list_find (conversions, s->charset);

        if (list == NULL)
                return;

#ifdef HAVE_ICONV
        translate (s, list->to);
#else
        error_ (0, _("conversions are unavailable on your system"));
#endif
}



static char *
get_file_name (char *header)
{
        char *seek = strstr (header, "name=");
        char *start;
        char *result;

        if (seek == NULL)
                seek = strstr (header, "NAME=");
  
        if (seek == NULL)
                return NULL;

        start = seek + 5;
        if (*start == '"')
                start++;

        seek = start;
        while (*seek != '"' && *seek != '\r' && *seek != '\n' && *seek
               && *seek != ';'){
                seek++;
        }
        result = xmalloc (seek - start + 1);
        result[seek - start] = '\0';
        memcpy (result, start, seek - start);

        return mime_decode_header (result, seek - start, 1);
}



static void
get_type (mime_t *mime, char *str)
{
        int         ret;
        int         len;
        regmatch_t  matches[5];

        if (! mime_type_compiled)
                return;

        ret = regexec (& mime_type_re, str, 5, matches, 0);

        if (ret && ret != REG_NOMATCH){
                error_regex (ret, & mime_type_re, NULL);
                return;
        }

        if (ret == REG_NOMATCH){
                error_ (0, _("invalid mime type '%s'"), str);
                return;
        }
                
        len        = matches[1].rm_eo - matches[1].rm_so;
        mime->type = xmalloc (len + 1);
        memcpy (mime->type, str + matches[0].rm_so, len);
        mime->type[len] = '\0';

        if (matches[4].rm_so == -1 || matches[4].rm_eo == -1)
                return;

        len           = matches[4].rm_eo - matches[4].rm_so;
        mime->charset = xmalloc (len + 1);
        memcpy (mime->charset, str + matches[4].rm_so, len);
        mime->charset[len] = '\0';
}



static void
destroy_s (struct rfc2047 *s)
{
        if (s->inter && s->result){
                str_destroy (s->inter);
        }
        else if (s->inter){
                s->result = s->inter;
        }

        if (s->charset)
                xfree (s->charset);
}



static void
determine_type (mime_t *mime)
{
        struct mime_handler *type = find_type (handlers, mime->file_name);

        if (type == NULL){
                mime->type     = xstrdup ("application/octet-stream");
                mime->encoding = MENC_BASE64;
        }
        else {
                mime->type = xstrdup (type->type);
                if (mime_is_text (mime))
                        mime->encoding = MENC_NONE;
                else
                        mime->encoding = MENC_BASE64;
        }
}



static void
register_new_handler (char *type, char *re, char *handler)
{
        int                  ret;
        struct mime_handler *new;
        
        new          = xmalloc (sizeof (struct mime_handler));
        new->type    = type;
        new->handler = handler;
        ret = regcomp (& new->re, re, REG_ICASE | REG_EXTENDED | REG_NEWLINE);

        if (ret){
                error_regex (ret, & new->re, re);
                destroy_handler (new);
                return;
        }

        new->next = handlers;
        handlers  = new;
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
mime_init (void)
{
        int ret;
        
        ret = regcomp (& encoded_string_re,
                       "(\n[ \t])?=\\?([^?]+)\\?([QB])\\?([^?]+)\\?=",
                       REG_ICASE | REG_EXTENDED | REG_NEWLINE);

        if (ret){
                encoded_compiled = 0;
                error_regex (ret, & encoded_string_re, "internal re");
                regfree (& encoded_string_re);
        }
        else {
                encoded_compiled = 1;
        }

        ret = regcomp (& mime_type_re,
                       "([a-z\\-]+(/[a-z0-9\\.\\-]+)?)(;[ \t\r\n]*charset=\"?([a-z0-9\\-]+)\"?)?",
                       REG_ICASE | REG_EXTENDED | REG_NEWLINE);

        if (ret){
                mime_type_compiled = 0;
                error_regex (ret, & mime_type_re, "internal re");
                regfree (& mime_type_re);
        }
        else {
                mime_type_compiled = 1;
        }
}



void
mime_post_setup (void)
{
        register_new_handler (xstrdup ("text/plain"),
                              "(\\.txt$)|(README)|(\\.diff$)|(\\.asc$)"
                              "|(\\.[ch]$)|(\\.cpp$)|(\\.cc$)|(\\.java$)"
                              "|(\\.sh$)|(\\.tex$)|(\\.pl$)|(\\.po$)",
                              xstrdup ("elmo"));
        register_new_handler (xstrdup ("text/html"),
                              "\\.[dsx]?html?$",
                              xstrdup ("elmo"));
        register_new_handler (xstrdup ("text/rtf"),
                              "\\.rtf$",
                              NULL);
        register_new_handler (xstrdup ("text/xml"),
                              "\\.x[ms]l$",
                              NULL);
        register_new_handler (xstrdup ("text/css"),
                              "\\.css$",
                              NULL);
        register_new_handler (xstrdup ("image/gif"),
                              "\\.gif$",
                              NULL);
        register_new_handler (xstrdup ("image/jpeg"),
                              "\\.jpe?g?$",
                              NULL);
        register_new_handler (xstrdup ("image/png"),
                              "\\.png$",
                              NULL);
        register_new_handler (xstrdup ("image/bmp"),
                              "\\.bmp$",
                              NULL);
        register_new_handler (xstrdup ("image/pcx"),
                              "\\.pcx$",
                              NULL);
        register_new_handler (xstrdup ("image/tiff"),
                              "\\.tiff?$",
                              NULL);
        register_new_handler (xstrdup ("audio/mpeg"),
                              "(\\.mp[23]$)|(\\.mpe?ga$)",
                              NULL);
        register_new_handler (xstrdup ("audio/midi"),
                              "\\.midi?$",
                              NULL);
        register_new_handler (xstrdup ("video/mpeg"),
                              "\\.mp[eg]g?$",
                              NULL);
}



void
mime_free_resources (void)
{
        if (encoded_compiled)
                regfree (& encoded_string_re);
        encoded_compiled = 0;

        if (mime_type_compiled)
                regfree (& mime_type_re);
        mime_type_compiled = 0;

        list_destroy (conversions);
        conversions = NULL;

        list_destroy (prepared);
        prepared = NULL;

        destroy_handler (handlers);
        handlers = NULL;
}



void
mime_add_charset (char *str)
{
        if (prepared){
                prepared->to   = str;
                prepared->next = conversions;
                conversions    = prepared;

                prepared       = NULL;
        }
        else {
                prepared       = xmalloc (sizeof (struct pair_list));
                prepared->from = str;
                prepared->to   = NULL;
                prepared->next = NULL;
        }
}



void
mime_register_handler (char *type, char *re, char *handler)
{
        int                  ret;
        struct mime_handler *new = find_mime (handlers, type);
        
        if (new){
                xfree (type);

                if (new->handler)
                        xfree (new->handler);
                regfree (& new->re);

                new->handler = handler;

                ret = regcomp (& new->re, re,
                               REG_ICASE | REG_EXTENDED | REG_NEWLINE);
                if (ret){
                        error_regex (ret, & new->re, re);
                }
        }
        else {
                register_new_handler (type, re, handler);
        }

        xfree (re);
}



int
mime_is_text (mime_t *mime)
{
        regmatch_t match[1];
        
        if (mime == NULL)
                return 0;
        if (mime->type == NULL)
                return 1;
        return misc_regex ("(^text/.*)|(^message/.*)", mime->type, match);
}



/* RFC2047 */
char *
mime_decode_header (char *str, int len, int free_s)
{
        int             ret;
        struct rfc2047  s;

        s.str      = str;
        s.inter    = NULL;
        s.result   = NULL;
        s.special_ = 1;
        
        ret = find_encoded_string (& s);
        if (! ret && free_s){
                return str;
        }
        else if (! ret){
                return xstrdup (str);
        }

        save_charset (& s);
        get_intermediate (& s);
        get_result (& s);
        
        if (free_s)
                xfree (str);

        destroy_s (& s);
        return str_finished (s.result);
}



str_t *
mime_decode (mime_t *mime, char *str, int free_s)
{
        int             len;
        str_t          *result = NULL;
        struct rfc2047  s;

        if (mime == NULL || str == NULL)
                return NULL;

        len = mime->off_end - mime->off_start;

        switch (mime->encoding){

                case MENC_NONE:
                case MENC_7BIT:
                case MENC_8BIT:
                        result = str_from_str (str, len);
                        break;

                case MENC_QP:
                        result = str_create_size (len + 1);
                        decode_qp (result, str, str + len - 1, 0);
                        break;

                case MENC_BASE64:
                        result = str_create_size (len + 1);
                        decode_base64 (result, str, str + len - 1);
                        break;

                case MENC_UUENCODE:
                        error_ (0, _("uudecode unimplemented yet"));
                        return str_from_str (str, len);

                default:
                        debug_msg (DEBUG_ERROR, "invalid encoding in mime_decode");
                        error_ (0, "%s", _("You have found a bug. Please submit a bug report"));
                        return NULL;
        }

        if (result->str != str && free_s)
                xfree (str);
        
        if (mime->charset == NULL)
                return result;

        s.charset = mime->charset;
        s.inter   = result;
        s.result  = NULL;

        get_result (& s);

        s.charset = NULL;

        destroy_s (& s);
        return s.result;
}



str_t *
mime_encode (mime_t *mime, char *bytes, int len)
{
        int    rlen;
        str_t *result;

        switch (mime->encoding){

                case MENC_NONE:
                case MENC_7BIT:
                case MENC_8BIT:
                        return str_from_str (bytes, len);

                case MENC_QP:
                        error_ (0, _("q-p encoding unimplemented yet"));
                        return str_from_str (bytes, len);

                case MENC_BASE64:
                        rlen   = (4 * len) / 3;
                        rlen  += rlen / 76 + 1;
                        rlen  += 5;
                        result = str_create_size (rlen + 1);
                        encode_base64 (result, bytes, bytes + len - 1);
                        xfree (bytes);
                        return result;

                case MENC_UUENCODE:
                        error_ (0, _("uuencode unimplemented yet"));
                        return str_from_str (bytes, len);

                default:
                        debug_msg (DEBUG_ERROR, "invalid encoding in mime_encode");
                        error_ (0, "%s", _("You have found a bug. Please submit a bug report"));
                        return NULL;
        }
}



mime_t *
mime_create (void)
{
        mime_t *mime;

        mime             = xmalloc (sizeof (mime_t));

        mime->encoding   = MENC_NONE;
        mime->type       = NULL;
        mime->charset    = NULL;

        mime->off_header = -1;
        mime->off_start  = -1;
        mime->off_end    = -1;
        mime->off_bound  = -1;

        mime->file_name  = NULL;

        return mime;
}



mime_t *
mime_from_file_name (const char *fname)
{
        mime_t *mime;

        mime            = mime_create ();
        mime->file_name = xstrdup (fname);

        determine_type (mime);
        
        return mime;
}



void
mime_destroy (mime_t *mime)
{
        if (mime->file_name)
                xfree (mime->file_name);
        if (mime->type)
                xfree (mime->type);
        if (mime->charset)
                xfree (mime->charset);
        xfree (mime);
}



void
mime_set_from_header (mime_t *mime, char *str)
{
        mime->file_name = get_file_name (str);

        get_type (mime, str);
}



void
mime_complete_file_name (mime_t *mime, char *str)
{
        if (mime->file_name)
                return;

        mime->file_name = get_file_name (str);
}



char *
mime_encoding_str (mime_t *mime)
{
        switch (mime->encoding){

                case MENC_NONE:
                        return NULL;
                
                case MENC_7BIT:
                        return "7bit";

                case MENC_8BIT:
                        return "8bit";
                
                case MENC_QP:
                        return "quoted-printable";

                case MENC_BASE64:
                        return "base64";

                case MENC_UUENCODE:
                        return "uuencode";
        }
        return NULL;
}



void
mime_dump (memchunk_t *memchunk, mime_t *mime)
{
        memchunk_intdump (memchunk, mime->encoding);
        memchunk_strdump (memchunk, mime->type);
        memchunk_strdump (memchunk, mime->charset);

        memchunk_intdump (memchunk, mime->off_header);
        memchunk_intdump (memchunk, mime->off_start);
        memchunk_intdump (memchunk, mime->off_end);
        memchunk_intdump (memchunk, mime->off_bound);

        memchunk_strdump (memchunk, mime->file_name);
}



mime_t *
mime_read (memchunk_t *memchunk)
{
        mime_t *mime;

        mime             = mime_create ();
        mime->encoding   = memchunk_intget (memchunk);
        mime->type       = memchunk_strget (memchunk);
        mime->charset    = memchunk_strget (memchunk);

        mime->off_header = memchunk_intget (memchunk);
        mime->off_start  = memchunk_intget (memchunk);
        mime->off_end    = memchunk_intget (memchunk);
        mime->off_bound  = memchunk_intget (memchunk);

        mime->file_name  = memchunk_strget (memchunk);
  
        return mime;
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE mime.c
 *
 ****************************************************************************/
